
/*
 * UNFS3 server framework
 * Originally generated using rpcgen
 * Portions (C) 2004, Pascal Schmidt <der.eremit@email.de>
 * see file LICENSE for license details
 */

#include <stdarg.h>
#include <rpc/rpc.h>
#include <rpc/pmap_clnt.h>
#include <netinet/in.h>
#include <ownfsd.h>

#if HAVE_RPC_SVC_SOC_H == 1
# include <rpc/svc_soc.h>
#endif

// #include "Config/exports.h"

#ifndef SIG_PF
#define SIG_PF void(*)(int)
#endif

#define UNFS_NAME "OWNFSD 1-wire NFS filesystem -- see http://owfs.sourceforge.net palfille@earthlink.net\n"

enum protocol_type { protocol_tcp, protocol_udp, } ;

static SVCXPRT *create_udp_transport( int port ) ;
static SVCXPRT *create_tcp_transport( int port ) ;
static void register_mount_service(SVCXPRT * udptransp, SVCXPRT * tcptransp) ;
static void register_nfs_service(SVCXPRT * udptransp, SVCXPRT * tcptransp) ;


static void register_nfs_service(SVCXPRT * udptransp, SVCXPRT * tcptransp) {
    u_long tcp = 0 ;
    u_long udp = 0 ;
    
    if (!NFS_params.no_portmapper) {
        pmap_unset(NFS_params.NFS_program, NFS_params.NFS_version);
        tcp = IPPROTO_TCP ;
        udp = IPPROTO_UDP ;
    }

    if (udptransp != NULL) {
        /* Register NFS service for UDP */
        if (!svc_register(udptransp, NFS_params.NFS_program, NFS_params.NFS_version, nfs3_program_3, udp)) {
            LEVEL_CONNECT("unable to register (NFS3_PROGRAM, NFS_V3, udp).")
            ow_exit(0);
        }
    }

    if (tcptransp != NULL) {
        /* Register NFS service for TCP */
        if (!svc_register(tcptransp, NFS_params.NFS_program, NFS_params.NFS_version, nfs3_program_3, tcp)) {
            LEVEL_CONNECT("unable to register (NFS3_PROGRAM, NFS_V3, tcp).")
            ow_exit(0);
        }
    }
}

static void register_mount_service(SVCXPRT * udptransp, SVCXPRT * tcptransp) {
    u_long tcp = 0 ;
    u_long udp = 0 ;

    if (!NFS_params.no_portmapper) {
        pmap_unset(MOUNTPROG, MOUNTVERS1);
        pmap_unset(MOUNTPROG, MOUNTVERS3);
        tcp = IPPROTO_TCP ;
        udp = IPPROTO_UDP ;
    }

    if (udptransp != NULL) {
        /* Register MOUNT service (v1) for UDP */
        if (!svc_register(udptransp, MOUNTPROG, MOUNTVERS1, mountprog_3, udp)) {
            LEVEL_CONNECT("unable to register (MOUNTPROG, MOUNTVERS1, udp).")
            ow_exit(0);
        }

        /* Register MOUNT service (v3) for UDP */
        if (!svc_register(udptransp, MOUNTPROG, MOUNTVERS3, mountprog_3, udp)) {
                LEVEL_CONNECT("unable to register (MOUNTPROG, MOUNTVERS3, udp).")
            ow_exit(0);
        }
    }

    if (tcptransp != NULL) {
        /* Register MOUNT service (v1) for TCP */
        if (!svc_register
            (tcptransp, MOUNTPROG, MOUNTVERS1, mountprog_3, tcp)) {
            LEVEL_CONNECT("unable to register (MOUNTPROG, MOUNTVERS1, tcp).")
            ow_exit(0);
        }

        /* Register MOUNT service (v3) for TCP */
        if (!svc_register
            (tcptransp, MOUNTPROG, MOUNTVERS3, mountprog_3, tcp)) {
            LEVEL_CONNECT("unable to register (MOUNTPROG, MOUNTVERS3, tcp).")
            ow_exit(0);
        }
    }
}

static SVCXPRT *create_udp_transport( int port ) {
    SVCXPRT * transp = NULL ;
    struct sockaddr_in sockin ;
    int sock = RPC_ANYSOCK ;
    const int on = 1 ;

    if ( port ) {
        sockin.sin_family      = AF_INET ;
        sockin.sin_port        = htons(port) ;
        sockin.sin_addr.s_addr = INADDR_ANY ;
        if ( (sock=socket(PF_INET, SOCK_DGRAM, 0)) == -1 ) {
            LEVEL_CONNECT("Cannot create udp socket on port %d.\n",port)
            return NULL ;
        }
        if ( setsockopt(sock,SOL_SOCKET, SO_REUSEADDR, (const char *)&on, sizeof(on) ) ) {
            LEVEL_CONNECT("Cannot set socket options for udp socket.\n")
            close(sock) ;
            return NULL ;
        }
        if ( bind(sock, (struct sockaddr *)&sockin, sizeof(struct sockaddr)) ) {
            LEVEL_CONNECT("Problen with bind on udp socket.\n")
            close(sock) ;
            return NULL ;
        }
    }
    if ( (transp=svcudp_create(sock)) ) {
        return transp ;
    }
    LEVEL_CONNECT("Error in svcudp_create.\n")
    close(sock) ;
    return NULL ;
}

static SVCXPRT *create_tcp_transport( int port ) {
    SVCXPRT * transp = NULL ;
    struct sockaddr_in sockin ;
    int sock = RPC_ANYSOCK ;
    const int on = 1 ;

    if ( port ) {
        sockin.sin_family      = AF_INET ;
        sockin.sin_port        = htons(port) ;
        sockin.sin_addr.s_addr = INADDR_ANY ;
        if ( (sock=socket(PF_INET, SOCK_STREAM, 0)) == -1 ) {
            LEVEL_CONNECT("Cannot create tcp socket on port %d.\n",port)
            return NULL ;
        }
        if ( setsockopt(sock,SOL_SOCKET, SO_REUSEADDR, (const char *)&on, sizeof(on) ) ) {
            LEVEL_CONNECT("Cannot set socket options for tcp socket.\n")
            close(sock) ;
            return NULL ;
        }
        if ( bind(sock, (struct sockaddr *)&sockin, sizeof(struct sockaddr)) ) {
            LEVEL_CONNECT("Problen with bind on tcp socket.\n")
            close(sock) ;
            return NULL ;
        }
    }
    if ( (transp=svctcp_create(sock,0,0)) ) {
        return transp ;
    }
    LEVEL_CONNECT("Error in svctcp_create.\n")
    close(sock) ;
    return NULL ;
}

void ow_exit( int e ) {
    if (e == SIGHUP || e == SIGUSR1) {
        return;
    }

    if (!NFS_params.no_portmapper) {
        svc_unregister(MOUNTPROG, MOUNTVERS1);
        svc_unregister(MOUNTPROG, MOUNTVERS3);
        svc_unregister(NFS_params.NFS_program,NFS_params.NFS_version);
    }

    if (e == SIGSEGV) LEVEL_DEFAULT("segmentation fault")

    LibClose() ;
    if( log_available) closelog() ;
    
    exit(e) ;
}

/*
 * NFSD main function
 * originally generated by rpcgen
 * forking, logging, options, and signal handler stuff added
 */
int main(int argc, char **argv) {
    int c ;
    register SVCXPRT *tcptransp = NULL;
    register SVCXPRT *udptransp = NULL;

    if ( argc>0 ) progname = strdup(argv[0]) ;

//    mtrace() ;
    /* Set up owlib */
    LibSetup() ;

    while ( (c=getopt_long(argc,argv,OWLIB_OPT,owopts_long,NULL)) != -1 ) {
        switch (c) {
            case 'V':
                fprintf(stderr,
                        "%s version:\n\t" VERSION "\n",argv[0] ) ;
                break ;
        }
        if ( owopt(c,optarg,opt_nfsd) ) ow_exit(0) ; /* rest of message */
    }

    while ( optind < argc-1 ) {
        OW_ArgGeneric(argv[optind]) ;
        ++optind ;
    }

    if ( indevices == 0 ) {
        LEVEL_DEFAULT("No 1-wire adapter specified.\n" )
        ow_exit(1) ;
    }

    if ( outdevices ) {
        LEVEL_DEFAULT("-p not valid for %s\n", progname )
        ow_exit(1) ;
    }

    /* NFS transports */
    if (!NFS_params.tcp_only) udptransp = create_udp_transport(NFS_params.NFS_port);
    if (!NFS_params.udp_only) tcptransp = create_tcp_transport(NFS_params.NFS_port);
    if ( udptransp==NULL && tcptransp==NULL ) {
        LEVEL_DEFAULT("No valid network protocol (tcp or udp) for NFS on port %d.\n",NFS_params.NFS_port)
        ow_exit(2) ;
    }   
    register_nfs_service(udptransp, tcptransp);

    /* MOUNT transports. If ports are equal, then the MOUNT service can reuse
       the NFS transports. */
    if (NFS_params.mount_port != NFS_params.NFS_port) {
        if (!NFS_params.tcp_only) udptransp = create_udp_transport(NFS_params.mount_port);
        if (!NFS_params.udp_only) tcptransp = create_tcp_transport(NFS_params.mount_port);
        if ( udptransp==NULL && tcptransp==NULL ) {
            LEVEL_DEFAULT("No valid network protocol (tcp or udp) for mount on port %d.\n",NFS_params.mount_port)
            ow_exit(2) ;
        }
    }
    register_mount_service(udptransp, tcptransp);

    /*
    * Now we drop privledges and become a daemon.
    */
    if ( LibStart() ) ow_exit(1) ;


    svc_run();
    exit(1);
    /* NOTREACHED */

    return 0;
}
